"""ratelimit.py

created 31-dec-2020 by richb@instantlinux.net
"""

from base64 import b64encode
from flask import g, request
import hashlib
import logging
from redis.exceptions import ConnectionError
import time

from .service_config import ServiceConfig
from .session_manager import SessionManager
from .grants import Grants

params = {}


class RateLimit(object):
    """Rate Limiting
    """
    def __init__(self, redis_conn=None):
        global params
        if not redis_conn:
            redis_conn = SessionManager().connection
        self.config = ServiceConfig().config
        if not params:
            params = dict(redis=redis_conn,
                          interval=self.config.RATELIMIT_INTERVAL)
        self.redis = params['redis']
        self.interval = params['interval']

    def call(self, service='*', uid=None):
        """Apply the granted rate limit for uid to a given service
        Two redis keys are used, for even and odd intervals as measured
        modulo the Unix epoch time. The current interval's key is
        incremented if we haven't reached the limit yet, then we delete
        that key at first call seen next interval and increment the other
        key. The keys expire within interval-1 seconds to keep the cache clear
        after the user stops sending calls.

        Args:
          service (str): name of a service
          uid (str): a user ID

        Returns:
          bool: true if limit exceeded
        """
        if not self.config.RATELIMIT_ENABLE:
            return False
        uid = uid or self._get_request_uid()
        if not uid:
            # No limit on anonymous requests, TODO make this smarter to
            #  protect against brute-force DDOS / security attacks
            return False
        limit = int(Grants().get('ratelimit', uid=uid))
        curr = round(time.time()) % (self.interval * 2) // self.interval
        key = 'rate:%s:%s:%d' % (service, uid, curr)
        altkey = 'rate:%s:%s:%d' % (service, uid, 0 if curr else 1)
        try:
            calls = self.redis.get(key)
        except ConnectionError as ex:
            logging.error(dict(action='ratelimit.call', message=str(ex)))
            return False
        if not calls or int(calls.decode()) < limit:
            try:
                pipe = self.redis.pipeline()
                pipe.incr(key).expire(
                    key, self.interval - 1).delete(altkey).execute()
            except Exception as ex:
                logging.error('action=ratelimit.call message=%s' % str(ex))
        else:
            logging.info(dict(action='ratelimit.call', uid=uid,
                              service=service, message='limit exceeded'))
            return True
        return False

    def _get_request_uid(self):
        """Examine headers to find uid
        """
        header_auth = self.config.HEADER_AUTH_APIKEY
        if header_auth in request.headers:
            try:
                prefix, secret = request.headers.get(header_auth).split('.')
            except ValueError:
                return None
            secret = b64encode(
                hashlib.sha1(secret.encode()).digest())[:8].decode('utf8')
            # TODO confirm whether this drops first use of apikey
            return g.session.get(None, secret, key_id=prefix, arg='sub')
        elif request.authorization:
            return request.authorization.username



"""storage.py

Storage access
  Files, photos and videos are stored as objects in a cloud vendor's
  storage service (Amazon S3 or equivalent). The backend server here
  acts as a broker for authenticating uploads and storing/retrieving
  metadata descriptors for each object. Metadata is stored in a
  redis cache.

created 25-jan-2020 by richb@instantlinux.net
"""

from b2sdk.v1 import B2Api, InMemoryAccountInfo
import base64
import boto3
from botocore.config import Config
from botocore.exceptions import ClientError
from datetime import datetime
from flask import g
import json
import logging
import re
import redis
from sqlalchemy.orm.exc import NoResultFound
import uuid

from ..access import AccessControl
from ..const import Constants
from ..grants import Grants
from ..service_config import ServiceConfig
import apicrud.utils as utils


class StorageAPI(object):
    """Storage API

    Attributes:
      credential_ttl (int): how long until temp credential expires
      redis_conn (obj): connection to redis
      uid (str): User ID
      db_session (obj): existing db session
    """

    def __init__(self, credential_ttl=None, redis_conn=None,
                 uid=None, db_session=None):
        config = ServiceConfig().config
        self.models = ServiceConfig().models
        self.redis_conn = (
            redis_conn or redis.Redis(host=config.REDIS_HOST,
                                      port=config.REDIS_PORT, db=0))
        self.cache_ttl = config.REDIS_TTL
        self.credential_ttl = credential_ttl or 86400
        self.uid = uid or AccessControl().uid
        self.db = db_session or g.db

    def get_upload_url(self, body):
        """Get a new pre-authorized URL from the storage vendor which
        allows the end user to upload one object to storage

        Args:
          body (dict):
            storage_id key provides ID in storage model; other keys provide
            metadata about the object to be uploaded, which will be stored
            in the new object's entry in database upon upload completion
        """
        storage_id = body.get('storage_id')
        try:
            storage = self.db.query(self.models.Storage).filter_by(
                id=storage_id, status='active').one()
        except NoResultFound:
            return dict(id=storage_id, message='storage volume not found'), 404
        resource = body.get('parent_type', 'album')
        if re.compile('^[a-f0-9]{8}-?[a-f0-9]{4}-?4[a-f0-9]{3}'
                      '-?[89ab][a-f0-9]{3}-?[a-f0-9]{12}.jpeg').match(
                          body.get('name').lower()):
            # Make iPhone filename shorter
            body['name'] = base64.urlsafe_b64encode(
                uuid.UUID(body['name'].strip('.jpeg')).bytes).rstrip(
                    b'=').decode('ascii') + '.jpeg'
        logmsg = dict(uid=self.uid, resource=resource, name=body.get('name'),
                      size=body.get('size'))

        model = dict(album=self.models.Album, list=self.models.List,
                     message=self.models.Message)[resource]
        if resource == 'album':
            try:
                record = self.db.query(model).filter_by(
                    id=body.get('parent_id'), status='active')
                album_size = len(record.one().pictures)
            except NoResultFound:
                return dict(album_id=body.get(
                    'parent_id'), message='album not found'), 404
            max_size = Grants(db_session=self.db).get('album_size')
            if album_size >= max_size:
                msg = 'album is full (max=%d)' % max_size
                logging.info(dict(message=msg, **logmsg))
                return dict(message=msg), 405
        if not AccessControl(model=model).with_permission(
                    'u', query=record):
            msg = 'access_denied'
            logging.warning(dict(message=msg, **logmsg))
            return dict(message='access denied'), 403

        self.vendor = storage.credentials.vendor
        id = utils.gen_id(prefix='f-')
        if body.get('content_type'):
            ctype = body['content_type'].split('/')[1]
        else:
            ctype = None
        storage_path = ('%s/%s/%s' % (
            storage.prefix if storage.prefix else '', self.uid, id)).strip('/')
        suffix = '.' + ctype if ctype else ''
        if ctype in Constants.MIME_VIDEO_TYPES:
            duration_max = Grants(
                db_session=self.db).get('video_duration_max')
            if body.get('duration') and body['duration'] > duration_max:
                msg = 'video exceeds maximum duration=%f' % duration_max
                logging.warning(dict(message=msg, **logmsg))
                return dict(message=msg), 405
        max_size = Grants(
            db_session=self.db).get('media_size_max')
        if body.get('size') and body['size'] > max_size:
            msg = 'file size exceeds max=%d' % max_size
            logging.info(dict(message=msg, **logmsg))
            return dict(message=msg), 405
        if self.vendor == 'backblaze':
            result, status = StorageBackblaze(
                self.models.Storage).get_upload_url(
                    storage_id, storage_path + suffix,
                    content_type=body.get('content_type'))
        elif self.vendor == 'aws':
            result, status = StorageS3(
                credentials=storage.credentials,
                credential_ttl=self.credential_ttl).get_upload_url(
                    storage.bucket, storage_path + suffix,
                    content_type=body.get('content_type'))
        else:
            return dict(message='Unsupported vendor=%s' % self.vendor), 405
        if status != 201:
            return result, status
        meta = dict(
            created=g.request_start_time.timestamp(),
            ctype=ctype,
            duration=body.get('duration'),
            fid=id,
            height=body.get('height'),
            modified=body.get('modified'),
            name=body.get('name'),
            oid=storage.uid,
            path=storage_path,
            pid=body.get('parent_id'),
            sid=storage_id,
            size=body.get('size'),
            uid=self.uid,
            width=body.get('width'))
        self.update_file_meta(id, meta)
        logging.info(dict(action='get_upload_url', **logmsg))
        return dict(params=result['fields'], upload_url=result['uploadUrl'],
                    file_id=id), status

    def get_file_meta(self, file_id):
        """Retrieve metadata from redis cache

        Args:
          file_id (str): ID in file model
        Returns:
          dict: fid, uid, sid, pid, path, name
        """
        key = 'fid:%s:%s' % (self.uid, file_id[-8:])
        try:
            meta = json.loads(self.redis_conn.get(key))
        except TypeError:
            return None
        except Exception as ex:
            logging.warn('action=get_upload_state uid=%s exception=%s' %
                         (self.uid, str(ex)))
            return None
        return meta

    def update_file_meta(self, file_id, meta):
        """Update metadata in redis cache

        Args:
          file_id (str): ID in file model
          meta (dict): keys to update
        """
        key = 'fid:%s:%s' % (self.uid, file_id[-8:])
        try:
            self.redis_conn.set(key, json.dumps(meta), ex=self.cache_ttl,
                                nx=True)
        except Exception as ex:
            logging.warn('action=update_file_meta uid=%s exception=%s' %
                         (self.uid, str(ex)))

    def del_file_meta(self, file_id, meta):
        """Remove a metadata item from cache

        Args:
          file_id (str): ID in file model
          meta (dict): keys to update
        """
        key = 'fid:%s:%s' % (self.uid, file_id[-8:])
        try:
            self.redis_conn.delete(key)
        except Exception as ex:
            logging.warn('action=del_file_meta uid=%s exception=%s' %
                         (self.uid, str(ex)))

    def upload_complete(self, file_id, status, func_worker):
        """Finalize upload - pass to the worker

        Args:
          file_id (str): ID of file meta in redis
          status (str): completion status from frontend ('done' if OK)
          func_worker (function): callback for passing to celery worker
        Returns:
          tuple: dict with file ID and message, and http status
        """

        meta = self.get_file_meta(file_id)
        if meta and status == 'done':
            duration = datetime.utcnow().timestamp() - meta['created']
            logging.info(dict(action='upload_complete', file_id=file_id,
                              duration='%.3f' % duration))
            func_worker(self.uid, file_id)
            return dict(file_id=file_id), 201
        else:
            msg = 'unhandled frontend status'
            logging.warning('action=upload_complete status=%s message=%s' %
                            (status, msg))
            return dict(file_id=file_id, message=msg), 405

    def fetch_album_meta(self, album_id, thumbnail_height):
        """Retrieve metadata of a media album in format suitable
        for react-image-gallery

        Args:
          album_id: ID in album model
          thumbnail_height (int): which scaled image to choose
        Returns:
          dict: a set of image descriptors suitable for display
        """

        try:
            album = self.db.query(self.models.Album).filter(
                self.models.Album.id == album_id,
                self.models.Picture.status == 'active').one()
        except NoResultFound:
            return []
        results = []
        for picture in album.pictures:
            orig_uri = None
            uri_path = '%s/%s' % (picture.storage.cdn_uri, picture.path)
            fmt = picture.format_original
            sizes = [int(x) for x in album.sizes.split(',')
                     if not picture.height or int(x) <= picture.height]
            if fmt in Constants.MIME_VIDEO_TYPES:
                # TODO see line 249 of xiaolin's app.js example
                logging.warning('resource=album action=get id=%s '
                                'message=videos_nyi' % picture.id)
                continue
            else:
                max_height = Grants(self.db).get('photo_res_max')
                if not picture.height or picture.height <= max_height:
                    orig_uri = '%s.%s' % (uri_path, fmt)
                else:
                    orig_uri = '%s.%d.%s' % (uri_path, max_height, fmt)
            results.append(dict(
                id=picture.id,
                description=picture.caption,
                duration=picture.duration,
                height=picture.height,
                imageSet=[dict(srcSet='%s.%d.%s' % (uri_path, res, fmt),
                               media='(max-width: %dpx)' % res) for
                          res in sizes],
                name=picture.name,
                original=orig_uri,
                path=picture.path,
                size=picture.size,
                thumbnail='%s.%d.%s' % (uri_path, thumbnail_height, fmt),
                type='%s/%s' % ('video' if fmt in Constants.MIME_VIDEO_TYPES
                                else 'image', fmt),
                uid=picture.uid,
                width=picture.width))
            if picture.height and picture.height <= max_height:
                results[-1]['imageSet'].append(dict(
                    srcSet='%s.%s' % (uri_path, fmt),
                    media='(max-width: %dpx)' % picture.height))
                results[-1]['imageSet'].append(dict(
                    srcSet='%s.%s' % (uri_path, fmt),
                    media='(min-width: %dpx)' % picture.height))
            else:
                results[-1]['imageSet'].append(dict(
                    srcSet='%s.%d.%s' % (uri_path, sizes[-1], fmt),
                    media='(min-width: %dpx)' % sizes[-1]))
        return results

    def get_object(self, storage_id, storage_key):
        """Get object from bucket

        Args:
          storage_id (str): ID in storage model
          storage_key (str): pathname within storage bucket
        """
        try:
            storage = self.db.query(self.models.Storage).filter_by(
                id=storage_id, status='active').one()
        except NoResultFound:
            logging.error('storage volume not found')
        if storage.credentials.vendor == 'aws':
            return StorageS3(credentials=storage.credentials).get_object(
                                 storage.bucket, storage_key)

    def put_object(self, storage_id, storage_key, content,
                   content_type='image/jpeg'):
        """Send a file to the storage bucket

        Args:
          storage_id (str): ID in storage model
          storage_key (str): pathname within storage bucket
          content (bytes): file content
          content_type (str): mime type
        """
        try:
            storage = self.db.query(self.models.Storage).filter_by(
                id=storage_id, status='active').one()
        except NoResultFound:
            logging.error('storage volume not found')
        if storage.credentials.vendor == 'aws':
            return StorageS3(credentials=storage.credentials).put_object(
                                 storage.bucket, storage_key, content,
                                 content_type=content_type)

    def del_object(self, storage_id, storage_key):
        """Remove a file from a storage bucket

        Args:
          storage_id (str): ID in storage model
          storage_key (str): pathname within storage bucket
        """
        try:
            storage = self.db.query(self.models.Storage).filter_by(
                id=storage_id, status='active').one()
        except NoResultFound:
            logging.error('storage volume not found')
        if storage.credentials.vendor == 'aws':
            return StorageS3(credentials=storage.credentials).del_object(
                                 storage.bucket, storage_key)


class StorageBackblaze(object):
    """Vendor-specific storage interface: Backblaze B2.
    Not yet implemented

    Args:
      model (obj): a storage model
      db_session (obj): existing db session
    """
    def __init__(self, model, db_session=None):
        self.api = B2Api(InMemoryAccountInfo())
        self.db = db_session
        self.model = model

    def get_upload_url(self, storage_id, storage_key,
                       content_type='image/jpeg'):
        """Get a presigned authorization URL for upload

        Args:
          storage_id (str): ID of storage model object
          storage_key (str): pathname of new object to store
          content_type (str): mime type
        Returns:
          tuple: dict with upload URL, http status
        """
        try:
            storage = self.db.query(self.model).filter_by(
                id=storage_id, status='active').one()
        except NoResultFound:
            return dict(id=storage_id, message='storage volume not found'), 404

        self.api.authorize_account(
            'production', storage.credentials.key, storage.credentials.secret)
        try:
            result = self.api.raw_api.get_upload_url(
                self.api.account_info.get_api_url(),
                self.api.account_info.get_account_auth_token(),
                storage.identifier)
        except Exception as ex:
            msg = str(ex)
            return dict(message=msg, id=storage_id), 405
        return result, 201


class StorageS3(object):
    """Vendor-specific storage interface: Amazon S3

    Args:
      credential_ttl (int): how long until temp credential expires
      credentials (obj): API key and secret
      access_key (str): API key (if credentials object not specified)
      secret_key (str): API secret (if credentials object not specified)
      region (str): data center region specification
    """
    def __init__(self, credential_ttl=3600, credentials=None, access_key=None,
                 secret_key=None, region=Constants.DEFAULT_AWS_REGION):
        self.credential_ttl = credential_ttl
        self.region = region
        if credentials:
            self.access_key = credentials.key
            self.secret_key = credentials.secret
        else:
            self.access_key = access_key
            self.secret_key = secret_key
        self.api = boto3.client(
            's3', aws_access_key_id=self.access_key,
            aws_secret_access_key=self.secret_key,
            region_name=self.region,
            # signature_version = 's3v4',
            config=Config(s3={'addressing_style': 'virtual'}))

    def get_upload_url(self, bucket, storage_key,
                       content_type='image/jpeg'):
        """Get a presigned authorization URL for upload

        Args:
          bucket (str): name of bucket
          storage_key (str): pathname of new object to store
          content_type (str): mime type
        Returns:
          tuple: dict with uploadUrl, http status
        """
        try:
            result = self.api.generate_presigned_post(
                bucket, storage_key, ExpiresIn=self.credential_ttl)
        except ClientError as ex:
            msg = str(ex)
            return dict(message=msg, bucket=bucket), 405
        return dict(fields=result['fields'], uploadUrl=result['url']), 201

    def get_object(self, bucket, storage_key):
        """Fetch the object into a byte array

        Args:
          bucket (str): name of bucket
          storage_key (str): pathname of new object to store
        Returns:
          bytes
        """
        return self.api.get_object(Bucket=bucket,
                                   Key=storage_key)['Body'].read()

    def put_object(self, bucket, storage_key, content,
                   content_type='image/jpeg'):
        """Store a byte-array object

        Args:
          bucket (str): name of bucket
          storage_key (str): pathname of new object to store
          content (bytes): object
          content_type (str): mime type
        Returns:
          tuple: vendor results
        """
        logging.debug(dict(action='put', bucket=bucket,
                           storage_key=storage_key, content_type=content_type))
        return self.api.put_object(
            Bucket=bucket, Key=storage_key, Body=content,
            ContentType=content_type)

    def del_object(self, bucket, storage_key):
        """Delete an object

        Args:
          bucket (str): name of bucket
          storage_key (str): pathname of new object to store
        Returns:
          tuple: vendor results
        """
        logging.debug(dict(action='del', bucket=bucket,
                           storage_key=storage_key))
        return self.api.delete_object(Bucket=bucket, Key=storage_key)



"""service_registry.py

created 8-may-2020 by richb@instantlinux.net
"""

import json
import logging
import socket
import threading

from .aes_encrypt import AESEncrypt
from .service_config import ServiceConfig
from .session_manager import SessionManager
from . import utils

refresh_thread = None
service_data = {}
params = {}


class ServiceRegistry(object):
    """Service Registry

    Services or the UI discover one another through this service
    registry.  Each microservice instance submits its identity and
    capabilities to this central registry, implemented as expiring redis
    keys which are updated at a fixed frequency. Encryption provides
    modest protection against injection attacks.

    Args:
      aes_secret (str): an AES secret [default: config.REDIS_AES_SECRET]
      public_url (str): URL to serve [default: config.PUBLIC_URL]
      redis_conn (obj): connection to redis
      reload_params (bool): force param reload, for unit-testing
      ttl (int): how long to cache instance's registration
    """

    def __init__(self, aes_secret=None, redis_conn=None, public_url=None,
                 reload_params=False, ttl=None):
        global params, refresh_thread
        self.config = config = ServiceConfig().config
        self.public_url = public_url or config.PUBLIC_URL
        if not redis_conn:
            redis_conn = SessionManager().connection
        if not params or reload_params:
            params = dict(
                aes=AESEncrypt(aes_secret or config.REDIS_AES_SECRET),
                interval=config.REGISTRY_INTERVAL,
                ttl=ttl or config.REGISTRY_TTL)
        params['connection'] = redis_conn
        if not refresh_thread:
            refresh_thread = threading.Thread()

    def register(self, resource_endpoints, service_name=None,
                 instance_id=socket.gethostname(), tcp_port=None):
        """register an instance serving a list of endpoints

        Args:
          resource_endpoints (list of str): controller endpoints served
          service_name (str): microservice name
          instance_id (str): unique ID of instance
          tcp_port (int): port number of service
        """

        global service_data
        try:
            ipv4 = socket.gethostbyname(instance_id)
        except socket.gaierror:
            ipv4 = None
        try:
            redis_ip = socket.gethostbyname(self.config.REDIS_HOST)
        except socket.gaierror:
            redis_ip = None
        service_data['info'] = dict(
            endpoints=resource_endpoints, ipv4=ipv4,
            port=tcp_port or self.config.APP_PORT, public_url=self.public_url,
            created=utils.utcnow().replace(microsecond=0).isoformat())
        service_data['name'] = service_name or self.config.SERVICE_NAME
        service_data['id'] = instance_id
        logging.info(dict(action='service.register', id=instance_id,
                          redis_ip=redis_ip,
                          name=service_data['name'], **service_data['info']))
        self._save_entry()
        refresh_thread = threading.Timer(1, ServiceRegistry.update, ())
        refresh_thread.start()

    @staticmethod
    def update():
        """background function to update registration at the defined
        interval from local memory cache, until the instance
        terminates.
        """

        ServiceRegistry._save_entry()
        refresh_thread = threading.Timer(params['interval'],
                                         ServiceRegistry.update, ())
        refresh_thread.start()

    @staticmethod
    def _save_entry():
        key = 'reg:%s:%s' % (service_data['name'], service_data['id'])
        try:
            params['connection'].set(key, params['aes'].encrypt(json.dumps(
                service_data['info'])), ex=params['ttl'], nx=False)
        except Exception as ex:
            logging.error('action=registry.update message=%s' % str(ex))

    def find(self, service_name=None):
        """ Finds one or all services

        Args:
          service_name (str): a service, or None for all

        Returns:
          dict - instances (list of registered services)
                 url_map (public url for each top-level resource)
        """

        key = 'reg:%s:*' % service_name if service_name else 'reg:*'
        retval = []
        url_map = {}
        for instance in params['connection'].scan_iter(match=key):
            try:
                info = json.loads(params['aes'].decrypt(
                    params['connection'].get(instance)))
                info['name'], info['id'] = instance.decode().split(':')[1:]
                retval.append(info)
            except (TypeError, json.JSONDecodeError, UnicodeDecodeError):
                # A decrypt failure probably means another cluster
                # is sharing the same redis instance (e.g. in a dev env)
                # so skip over those service instances
                continue
            except Exception as ex:
                logging.error('action=registry.find key=%s exception=%s' %
                              (instance, str(ex)))
                continue
            if not info.get('endpoints'):
                continue
            base_url = ServiceConfig().config.BASE_URL
            for resource in info['endpoints']:
                if resource not in url_map:
                    url_map[resource] = info['public_url'] + base_url
                elif url_map[resource] != info['public_url'] + base_url:
                    logging.warning(dict(
                        action='registry.find', resource=resource,
                        message='misconfigured cluster serves two URLs',
                        instance1=':'.join(instance.decode().split(':')[1:]),
                        url1=info['public_url'] + base_url,
                        url2=url_map[resource]))
        return dict(instances=retval, url_map=url_map)



"""access.py

created 20-may-2019 by richb@instantlinux.net
refactored 6-mar-2020
"""

from base64 import b64encode
from datetime import timedelta
from flask import g, request
from flask_babel import _
import hashlib
import logging
import re
from sqlalchemy import or_
from sqlalchemy.orm.exc import NoResultFound
import yaml

from .exceptions import APIcrudGrantsError
from .service_config import ServiceConfig
from .utils import gen_id, utcnow

LEVELS = {}
POLICIES = {}
PRIV_RES = {}


class AccessControl(object):
    """Role-based access control

    Definitions:

    - principal: a user or role
    - membership: parent resource type for privacy sharing
    - model: database model name (e.g. Person)
    - resource: resource type (e.g. person)
    - rbac: role-based access control (defined in rbac.yaml)
    - role: a group name (e.g. admin or list-<id>-<level>)
    - privacy: sharing options as defined in rbac.yaml (e.g. \
      secret [default], public, invitee, member, manager)
    - actions: crudlghij (create, read, update, del, list, guest/member, \
      host/manager, invitee, join)

    In rbac.yaml, define the RBAC policies for each principal/resource
    combination. That file will be parsed into a singleton variable upon
    initial startup.  This implementation implements RBAC similar to that of
    kubernetes or AWS IAM, with the added capability of a simple privacy
    permission within each object (database record) which creates an implied
    ACL for read-only access by members of the object's group.

    Group names currently used are:

    - admin
    - user
    - pending
    - person
    - <resource>-<id>-<privacy>

    These are defined in session_auth.py's account_login() method.

    Args:
      policy_file (str): name of the yaml definitions file
      model (obj): a model to be validated for permissions
    """
    def __init__(self, policy_file=None, model=None):
        if POLICIES and request:
            self.policies = POLICIES['policies']
            self.privacy_levels = LEVELS['levels']
            self.private_res = PRIV_RES['private_res']
            self.model = model
            self.models = ServiceConfig().models
            self.resource = model.__name__.lower() if model else None
            self.auth = None
            header_auth = ServiceConfig().config.HEADER_AUTH_APIKEY
            if header_auth in request.headers:
                # TODO: add support for HMAC request signing in place
                # of this plain-text X-Api-Key header method
                prefix, secret = request.headers.get(header_auth).split('.')
                item = g.session.get(None, self.apikey_hash(secret)[:8],
                                     key_id=prefix)
                if item:
                    self.auth = item.get('auth').split(':')
                    self.uid = uid = item.get('sub')
            elif request.authorization:
                uid = self.uid = request.authorization.username
                secret = request.authorization.password
                try:
                    self.auth = g.session.get(
                        uid, secret, arg='auth').split(':')
                except AttributeError:
                    pass
            self.primary_res = self.private_res[0]['resource']
            self.auth_ids = {self.primary_res: []}
            if self.auth:
                for role in self.auth:
                    ev = self._parse_id(role, self.primary_res)
                    if ev:
                        self.auth_ids[self.primary_res].append(ev)
                self.account_id = g.session.get(uid, secret, 'acc')
                self.identity = g.session.get(uid, secret, 'identity')
            else:
                # For anonymous-access paths that don't require security
                self.uid = self.account_id = self.identity = None

    def load_rbac(self, filename):
        """ Read RBAC default policies from rbac.yaml, process any
        string substitutions, and convert * for re.match()

        Args:
          filename (str): filename containing RBAC definitions
        """

        with open(filename, 'rt', encoding='utf8') as f:
            rbac = yaml.safe_load(f)
            self.policies = {}
        for res, policy in rbac['policies'].items():
            self.policies[res] = []
            for item in policy:
                self.policies[res].append(dict(
                    principal=item['principal'].format(
                    ).replace('*', '.*'),
                    resource=item['resource'].format(
                        resource=res).replace('*', '.*'),
                    actions=set(list(item['actions']))))
                self.privacy_levels = rbac['privacy_levels']
                self.private_res = rbac['private_resources']
                LEVELS['levels'] = self.privacy_levels
                POLICIES['policies'] = self.policies
                PRIV_RES['private_res'] = self.private_res

    def with_permission(self, access, query=None, new_uid=None,
                        membership=None, id=None):
        """Evaluate permission to access an object identified by
        an open query or new uid. Pass in at least one of the
        query/uid/eid params

        Args:
          access (str): one of lrwcd (list, read, write, create, delete)
          query (obj): a resource query by id in SQLalchemy
          new_uid (str): user id of a new record
          membership (str): resource type which defines membership privacy
          id (str): resource ID

        Returns:
          bool: True if access allowed
        """
        rbac = self.rbac_permissions(query=query, owner_uid=new_uid,
                                     membership=membership, id=id)
        if access in rbac:
            return True
        else:
            duration = (utcnow().timestamp() -
                        g.request_start_time.timestamp())
            logging.info(dict(
                action='with_permission', message=_('access denied'),
                resource=self.resource, uid=self.uid, ident=self.identity,
                access=access, rbac=rbac, duration='%.3f' % duration))
        return False

    def with_filter(self, query, access='r'):
        """Apply RBAC and privacy to a query

        Args:
          query (obj): a resource query in SQLalchemy
          access (str): one of lrwcd (list, read, write, create, delete)

        Returns:
          obj: updated SQLalchemy query with filter applied

        TODO restrictions on contact-read by list-id
        """
        if hasattr(self.model, 'privacy'):
            conditions = [self.model.privacy == 'public']
        else:
            conditions = []
        for policy in self.policies[self.resource]:
            if policy['principal'] == '.*' and access in policy['actions']:
                return query
            if self.auth:
                if (policy['resource'] == '%s:.*' % self.resource and
                        policy['principal'] in self.auth and
                        access in policy['actions']):
                    return query
                elif (policy['resource'] == '%s:{uid}' % self.resource and
                      policy['principal'].split('/')[0] in ('person', 'user')
                      and policy['principal'].split('/')[0] in self.auth and
                      access in policy['actions']):
                    conditions.append(self.model.uid == self.uid)
                    # Special case: add referrer too
                    if self.resource == 'contact':
                        query = query.join(self.models.Person)
                        conditions.append(
                            self.models.Person.referrer_id == self.uid)
                elif (policy['resource'] == '%s:{uid}' % self.resource and
                      access in policy['actions']):
                    for role in self.auth:
                        if re.match(policy['principal'], role):
                            conditions.append(self.model.uid == self.uid)
                            break
        return query.filter(or_(*conditions))

    def rbac_permissions(self, query=None, owner_uid=None, membership=None,
                         id=None, privacy=None):
        """Evaluate an access request for self.auth roles of
        self.uid in self.resource against defined policies

        Args:
          query (obj):  an existing record (takes precedence over owner_uid)
          owner_uid (str): owner-uid of a record
          membership (str): resource type which defines membership privacy
          id (str): the resource ID if membership is set

        Returns:
          set: actions available to principal
        """
        if query:
            try:
                record = query.one()
                if hasattr(record, 'uid'):
                    owner_uid = record.uid
                elif hasattr(record, 'referrer_id'):
                    owner_uid = record.referrer_id or record.id
                if hasattr(record, 'privacy'):
                    privacy = record.privacy
                if hasattr(record, self.private_res[0]['attr']):
                    membership = self.private_res[0]['resource']
                    id = getattr(record, self.private_res[0]['attr'])
            except NoResultFound as ex:
                logging.warning('action=evaluate_rbac message=%s' % str(ex))
        if not id:
            membership = None

        # TODO improve schema or yaml syntax to handle these special-cases
        # TODO make the contact find function return correct rbac
        deny_delete = set()
        if query and self.resource == 'contact':
            if self.uid and self.uid != owner_uid:
                owner_uid = query.one().owner.referrer_id
            # TODO redesign fragile dependency on primary contact
            try:
                g.db.query(self.models.Person).filter_by(
                    identity=query.one().info).one()
                deny_delete = set('d')
            except NoResultFound:
                pass

        actions, defaults = (set(), set())
        if privacy == 'public' or (
                self.auth and id and '%s-%s-%s' % (membership, id, privacy)
                in self.auth):
            defaults = set('r')

        for policy in self.policies[self.resource]:
            if policy['principal'] == '.*':
                defaults = set(list(policy['actions']))
            elif self.auth:
                for role in self.auth:
                    if role in ('person', 'user'):
                        role += '/{uid}'
                    match_id = None
                    # TODO match_id is a bit hacky, should format str
                    if 'uid' in policy['resource']:
                        match_id = self.uid
                    elif 'eid' in policy['resource']:
                        match_id = id

                    # TODO fix the wildcard matching
                    if (policy['principal'].format(eid=id, list_id=id,
                                                   uid=owner_uid) ==
                        role.format(eid=id, list_id=id, uid=self.uid) and
                        re.match(
                            '^%s$' % policy['resource'].format(
                                eid=id, list_id=id, uid=owner_uid),
                            '%s:%s' % (self.resource, match_id))):
                        actions |= policy['actions']

                    # special-case for person/contact principal:<res>-*
                    elif (policy['principal'] == '%s-*' % membership and
                          role.split('-')[0] == membership and
                          re.match(
                              '^%s$' % policy['resource'].format(
                                  eid=id, list_id=id, uid=owner_uid),
                              '%s:%s' % (self.resource, match_id))):
                        actions |= policy['actions']
        actions -= deny_delete
        return actions if len(actions) else defaults

    def apikey_create(self, max_keys=1):
        """Check that user hasn't exceeded max grant for API keys, then
        generate an API key - a 41-byte string. First 8 characters (48
        bits) are an access key ID prefix; last 32 characters (192
        bits) are the secret key.

        Args:
          max_keys (int) - maximum number of keys granted for current uid
        Returns: tuple
          key ID (str) - public portion of key
          secret (str) - secret portion
          hashvalue (str) - hash value for database
        Raises: APIcrudGrantsError
        """
        logmsg = dict(action='create', resource='apikey', uid=self.uid)
        if g.db.query(self.models.APIkey).filter_by(
                uid=self.uid).count() >= max_keys:
            msg = _(u'max allowed API keys exceeded')
            logging.warning(dict(message=msg, allowed=max_keys, **logmsg))
            raise APIcrudGrantsError(msg)
        secret = gen_id(length=35, prefix='')[-32:]
        key_id = gen_id(prefix='')
        return key_id, secret, self.apikey_hash(secret)

    def apikey_verify(self, key_id, secret):
        """Verify an API key

        Args:
          key_id (str): the public key_id at time of generation
          secret (str): the unhashed secret

        Returns: tuple
          uid (str): User ID if valid
          scopes (list): list of scope IDs
        """
        try:
            record = g.db.query(self.models.APIkey).filter_by(
                prefix=key_id, hashvalue=self.apikey_hash(secret),
                status='active').one()
            if not record.last_used or (
                    record.last_used + timedelta(hours=6) < utcnow()
                    and record.expires < utcnow()):
                record.last_used = utcnow()
                g.db.commit()
        except NoResultFound:
            logging.info(dict(action='api_key', key_id=key_id,
                              message='not found'))
            return None, None
        except Exception as ex:
            logging.error(dict(action='api_key', message=str(ex)))
            return None, None
        if record.expires and record.expires < utcnow():
            logging.info(dict(action='api_key', key_id=key_id,
                              uid=record.uid, message='expired'))
            return None, None
        return record.uid, record.scopes

    @staticmethod
    def apikey_hash(secret):
        """Generate a hash value from the secret
        Args:
          secret (str): secret key
        """
        return b64encode(
            hashlib.sha256(secret.encode()).digest()).decode('utf8')

    @staticmethod
    def _parse_id(role, resource):
        """Given role <res>-<id>-<level>
        Returns id if resource matches, otherwise None

        Args:
          role (str): assigned role
          resource (str): resource type
        """
        x = role.split('-')
        if x and x[0] == resource:
            return '-'.join(x[1:-1])



"""grants.py

created 27-may-2019 by richb@instantlinux.net
"""

from flask import g
from datetime import timedelta
import json

from .access import AccessControl
from .service_config import ServiceConfig
from . import utils

GRANTS = {}


class Grants(object):
    """Account usage limits

    An account's usage limits are specified here in the grants table;
    the free-service tier is defined and passed in via load_defaults().
    Records in grants table are owned by administrator-level user.
    If a record matches a user uid, the default grant name=value is
    overridden.

    Attributes:
      db_session (obj): existing db session
      ttl (int): how long to cache a grant in memory
    """

    def __init__(self, db_session=None, ttl=None):
        self.ttl = ttl or ServiceConfig().config.REDIS_TTL
        self.models = ServiceConfig().models
        try:
            self.session = db_session or g.db
        except RuntimeError as ex:
            if 'Working outside of application context' not in str(ex):
                raise

    def get(self, name, uid=None):
        """Get the cached value of a named grant, if it hasn't expired
        Note that if any grant assigned to a uid expires before
        others, the earliest expiration applies to all the uid's grants

        Args:
          name (str): name of a grant, as defined in service config
          uid (str): user ID
        Returns:
          str: granted limit
        """
        if name not in GRANTS['defaults']:
            raise AttributeError('Unknown grant %s' % name)
        if not uid:
            uid = AccessControl().uid
        if uid not in GRANTS or utils.utcnow() > GRANTS[uid]['expires']:
            records = self.session.query(self.models.Grant).filter_by(
                uid=uid).all()
            grants = dict(expires=utils.utcnow() + timedelta(seconds=self.ttl))
            for record in records:
                grants[record.name] = record.value
                if record.expires:
                    grants['expires'] = min(record.expires, grants['expires'])
            GRANTS[uid] = grants
        if name in GRANTS[uid] and utils.utcnow() < GRANTS[uid]['expires']:
            return GRANTS[uid][name]
        return GRANTS['defaults'][name]

    def crud_get(self, crud_results, id):
        """Process results from BasicCRUD.get() for grants endpoint.
        If the id is found in database, perform the standard CRUD
        get(). Otherwise, look for a hybrid id in form uid:grant and
        return the cached Grant value.

        Args:
            crud_results (tuple): preliminary response
            name (str): name filter, if specified
        """
        if crud_results[1] == 200 or ':' not in id:
            return crud_results
        acc = AccessControl(model=self.models.Grant)
        uid, grantname = id.split(':')
        rbac = ''.join(sorted(list(acc.rbac_permissions(owner_uid=uid))))
        if 'r' not in rbac:
            return dict('access denied'), 403
        return dict(id=id,  uid=uid, name=grantname, value=str(self.get(
            grantname, uid=uid)), rbac=rbac, status='active'), 200

    def find(self, crud_results, **kwargs):
        """Process results from BasicCRUD.find() for grants endpoint

        Args:
            crud_results (tuple): preliminary response
            name (str): name filter, if specified
        """
        filter = json.loads(kwargs.get('filter', '{}'))
        name = kwargs.get('name') or filter.get('name')
        acc = AccessControl(model=self.models.Grant)
        uid = filter.get('uid') or acc.uid
        rbac = ''.join(sorted(list(acc.rbac_permissions(owner_uid=uid))))
        result = []
        for key, val in GRANTS['defaults'].items():
            if name and key != name:
                continue
            for row in crud_results[0]['items']:
                if row['name'] == key:
                    result.append(row)
                    break
            else:
                result.append(dict(id='%s:%s' % (uid, key), uid=uid,
                                   name=key, value=str(val), rbac=rbac,
                                   status='active'))
        return dict(items=result, count=len(result)), crud_results[1]

    def uncache(self, uid):
        """Remove grants from cache, any time a user's status changes

        Args:
          uid (str): user ID
        """
        GRANTS.pop(uid, None)

    def load_defaults(self, defaults):
        """Load default values from a dict of keyword: value pairs

        Args:
          defaults (dict): new defaults
        """
        GRANTS['defaults'] = defaults



from .access import AccessControl
from .account_settings import AccountSettings
from .aes_encrypt import AESEncrypt
from .basic_crud import BasicCRUD
from .exceptions import *  # noqa
from .grants import Grants
from .ratelimit import RateLimit
from .service_config import ServiceConfig
from .service_registry import ServiceRegistry
from .session_auth import SessionAuth
from .session_manager import SessionManager, Mutex

__all__ = ('AccessControl', 'AccountSettings', 'AESEncrypt', 'BasicCRUD',
           'Grants', 'Mutex', 'RateLimit', 'ServiceConfig', 'ServiceRegistry',
           'SessionAuth', 'SessionManager')



# Metrics at scale are a hard problem; see
#  https://www.infoworld.com/article/3230455/how-to-use-redis-for-real-time-metering-applications.html
# The main counting is done in memory with redis but data has to be
#  periodically persisted here--and older entries have to be pruned,
#  which is an expensive delete operation.

# Food for thought: why not create (upon first use for a given
# user/metric) a decrementing redis key that expires at the end of the
# grant period? If the redis deployment is sufficiently reliable,
# persistence to disk here might be unwanted. At worst, a redis crash
# would cause users to get free service for a while--a lower cost than
# trying to scale this damn thing.

# This was the metrics table I contemplated for sqlalchemy:

# class Metric(Base):
#     __tablename__ = 'metrics'
#     __table_args__ = (
#         UniqueConstraint(u'uid', u'name', u'day', name='uniq_metric'),
#     )
#
#     id = Column(String(16), primary_key=True, unique=True)
#     name = Column(Enum(u'direct', u'email', u'events', u'photos', u'sms'),
#                   nullable=False)
#     value = Column(BIGINT, nullable=False, server_default="0")
#     day = Column(INTEGER)
#     uid = Column(ForeignKey(u'people.id'), nullable=False, index=True)
#     created = Column(TIMESTAMP, nullable=False, server_default=func.now())
#     modified = Column(TIMESTAMP)
#
#     owner = relationship('Person', foreign_keys=[uid])



"""initialize.py

created 9-jan-2021 by richb@instantlinux.net
"""
import connexion
from flask import jsonify
from flask_cors import CORS
import logging
import os

from . import grants
from .access import AccessControl
from .const import Constants
from .service_config import ServiceConfig


def app(application):
    """Initialize the Flask app defined by openapi.yaml

    Args:
      application (obj): a connexion object
      init_func (function): any other function to call

    Returns:
      obj: Flask app
    """
    global babel

    config = ServiceConfig().config
    logging.basicConfig(level=config.LOG_LEVEL,
                        format='%(asctime)s %(levelname)s %(message)s',
                        datefmt='%m-%d %H:%M:%S')
    logging.getLogger('werkzeug').setLevel(logging.ERROR)
    application.app.config.from_object(config)
    application.add_api(
        os.path.join(os.path.dirname(os.path.abspath(__file__)),
                     Constants.SERVICE_CONFIG_FILE), base_path='/config/v1')
    application.add_api(config.OPENAPI_FILE)
    application.add_error_handler(400, render_status_4xx)
    application.add_error_handler(429, render_status_4xx)
    application.add_error_handler(connexion.ProblemException,
                                  render_problem)
    CORS(application.app,
         resources={r"/api/*": {'origins': config.CORS_ORIGINS}},
         supports_credentials=True)
    AccessControl().load_rbac(config.RBAC_FILE)
    grants.Grants().load_defaults(config.DEFAULT_GRANTS)
    logging.info(dict(action='initialize_app', port=config.APP_PORT))
    return application.app


def render_status_4xx(error):
    """Render function to provide message in dict as required for
    react-admin to display text of message for 4xx error codes

    Args:
      error (obj): the error object with name and description
    """
    return jsonify(dict(
        message=error.description,
        error=dict(status=error.name, code=error.code))), error.code


def render_problem(error):
    """Render function to provide message in dict as required for
    react-admin to display text of message for exception
    connexion.ProblemException

    Args:
      error (obj): the error object with name and description
    """
    return jsonify(dict(
        message=error.detail,
        error=dict(status=error.title, code=error.status))), error.status



"""worker_processing.py

Media worker functions to process media
  Uploaded videos and photos contain metadata fields that we can parse
  and store in the database (plus the redis cache). Various sizes of
  images need to be sent to the storage API for quick retrieval; these
  functions are the background engine for such longer-running tasks.

created 3-feb-2020 by richb@instantlinux.net
"""

import celery
from datetime import datetime
import hashlib
import io
import logging
from PIL import Image
from PIL.ExifTags import GPSTAGS, TAGS

from ..account_settings import AccountSettings
from ..database import get_session
from ..service_config import ServiceConfig
from .storage import StorageAPI


class MediaUploadException(Exception):
    pass


class MediaProcessing(object):
    """Media Processing

    Attributes:
      uid (str): User ID
      file_id (str): ID of record in File model
      db_session (obj): database session
    """

    def __init__(self, uid, file_id, db_session=None):
        self.config = config = ServiceConfig().config
        self.models = ServiceConfig().models
        self.api = StorageAPI(uid=uid, db_session=db_session)
        self.db_session = db_session or get_session(
            scopefunc=celery.utils.threads.get_ident, db_url=config.DB_URL)
        self.file_id = file_id
        self.meta = self.api.get_file_meta(file_id)

    def __del__(self):
        self.api.del_file_meta(self.file_id, self.meta)
        self.db_session.remove()

    def photo(self, uid, meta):
        """metadata and scaling for still images

        Args:
          uid (str): User ID
          meta (dict): Image metadata
        """

        def _exif_get(field, max_width):
            if exif.get(field):
                return exif.get(field)[:max_width]

        # fetch the image
        suffix = "." + meta["ctype"]
        raw = self.api.get_object(meta["sid"], meta["path"] + suffix)
        if not raw:
            raise MediaUploadException("get_object failed for path=%s" %
                                       meta["path"])
        img = Image.open(io.BytesIO(raw))

        # parse its EXIF fields
        exif_raw = (hasattr(img, "_getexif") and img._getexif()) or {}
        exif = {TAGS[key]: val for key, val in exif_raw.items()
                if key in TAGS}
        geolat, geolong, altitude = None, None, None
        if exif.get("GPSInfo"):
            geolat, geolong, altitude = self._gpsexif(exif)

        # generate the 50-pixel thumbnail
        thumbnail = img.copy()
        thumbnail.thumbnail((self.config.THUMBNAIL_TINY,
                             self.config.THUMBNAIL_TINY), Image.ANTIALIAS)
        tbytes = io.BytesIO()
        thumbnail.save(tbytes, meta["ctype"])

        # construct and save a pictures db record
        try:
            if meta["modified"]:
                modified = datetime.strptime(meta["modified"],
                                             "%Y-%m-%dT%H:%M:%S.%fZ")
        except ValueError:
            logging.warn("action=photo message=invalid_date id=%s "
                         "modified=%s" % (meta["fid"], meta["modified"]))
        record = self.models.Picture(
            id=meta["fid"],
            name=meta["name"],
            path=meta["path"],
            uid=uid,
            category_id=AccountSettings(None, db_session=self.db_session,
                                        uid=meta["oid"]).get.category_id,
            compression=_exif_get("Compression", 16),
            datetime_original=exif.get("DateTimeOriginal") or modified,
            format_original=meta["ctype"],
            geolat=geolat,
            geolong=geolong,
            gps_altitude=altitude,
            height=img.height,
            is_encrypted=False,
            make=_exif_get("Make", 16),
            model=_exif_get("Model", 32),
            orientation=exif.get("Orientation"),
            privacy=u"invitee",
            sha1=hashlib.sha1(raw).digest(),
            sha256=hashlib.sha256(raw).digest(),
            thumbnail50x50=tbytes.getvalue(),
            size=meta["size"],
            status="active",
            storage_id=meta["sid"],
            width=img.width,
        )
        self.db_session.add(record)

        # add picture to album; first added is initial cover picture
        album = self.db_session.query(self.models.Album).filter_by(
            id=meta["pid"]).one()
        record = self.models.AlbumContent(
            album_id=meta["pid"],
            picture_id=meta["fid"],
            rank=len(album.pictures) + 1)
        self.db_session.add(record)
        if len(album.pictures) == 0 or album.cover_id is None:
            album.cover_id = meta["fid"]
            self.db_session.add(album)
        self.db_session.commit()

        # generate additional scaled sizes
        sizes = [int(x) for x in album.sizes.split(",")]
        for alt_size in sizes:
            if alt_size >= img.height:
                logging.info("skipping image name=%s resize=%d original=%d" %
                             (meta["name"], alt_size, img.height))
                continue
            width = int(img.width * alt_size / img.height)
            logging.info("generating image name=%s width=%d height=%d id=%s"
                         % (meta["name"], width, alt_size, meta["fid"]))
            alt_img = img.resize((width, alt_size), Image.ANTIALIAS)
            tbytes = io.BytesIO()
            alt_img.save(tbytes, meta["ctype"])
            if not self.api.put_object(
                    meta["sid"], "%s.%d%s" % (
                        meta["path"], alt_size, suffix),
                    tbytes.getvalue(),
                    content_type="image/%s" % meta["ctype"]):
                raise MediaUploadException("put_object failed")

        # stay in budget--throw away originals larger than max
        # TODO grants needs to work under celery
        # res_max = Grants(db_session=self.db_session,
        #                  ttl=config.CACHE_TTL).get('photo_res_max')
        res_max = self.config.DEFAULT_GRANTS.get('photo_res_max')
        if img.height > res_max:
            self.api.del_object(meta["sid"], meta["path"] + suffix)

        # TODO (eventually)-apply encryption when gallery frontend has support
        #  and strip exif tags

    def video(self, uid, meta):
        """metadata for videos - construct and save a pictures db record

        Args:
          uid (str): User ID
          meta (dict): Video metadata
        """
        # TODO decide whether to bother with exif or thumbnail

        try:
            if meta["modified"]:
                modified = datetime.strptime(meta["modified"],
                                             "%Y-%m-%dT%H:%M:%S.%fZ")
        except ValueError:
            logging.warn("action=video message=invalid_date id=%s "
                         "modified=%s" % (meta["fid"], meta["modified"]))
        record = self.models.Picture(
            id=meta["fid"],
            name=meta["name"],
            path=meta["path"],
            uid=uid,
            category_id=AccountSettings(None, db_session=self.db_session,
                                        uid=meta["oid"]).get.category_id,
            compression=None,
            datetime_original=modified,
            duration=meta["duration"],
            format_original=meta["ctype"],
            geolat=None,
            geolong=None,
            gps_altitude=None,
            height=meta["height"],
            is_encrypted=False,
            make=None,
            model=None,
            orientation=None,
            privacy=u"invitee",
            sha1=None,
            sha256=None,
            thumbnail50x50=None,
            size=meta["size"],
            status="active",
            storage_id=meta["sid"],
            width=meta["width"],
        )
        self.db_session.add(record)

        # add to album
        album = self.db_session.query(self.models.Album).filter_by(
            id=meta["pid"]).one()
        # TODO query AlbumContents max(rank) to handle edge case
        record = self.models.AlbumContent(
            album_id=meta["pid"],
            picture_id=meta["fid"],
            rank=len(album.pictures) + 1)
        self.db_session.add(record)
        self.db_session.commit()

    def _gpsexif(self, exif):
        """Convert GPS exif data to fixed-decimal format. For some
        reason there"s no good standard python library for doing this.

        Returns:
          tuple (lat, long, alt):
            first two are converted to fixed-precision int to fit in 32 bits
        """

        for item, val in GPSTAGS.items():
            if item in exif["GPSInfo"]:
                exif[val] = exif["GPSInfo"][item]

        # convert lat/long to fixed-precision 32-bit integers for SQL
        geolat, geolong, altitude = None, None, None
        if exif.get("GPSLatitude"):
            geolat = self._decimal_from_degrees(
                exif["GPSLatitude"], exif["GPSLatitudeRef"])
        if exif.get("GPSLongitude"):
            geolong = self._decimal_from_degrees(
                exif["GPSLongitude"], exif["GPSLongitudeRef"])
        if exif.get("GPSAltitude"):
            altitude = exif["GPSAltitude"][0] / exif["GPSAltitude"][1]
        return geolat, geolong, altitude

    @staticmethod
    def _decimal_from_degrees(dms, hemisphere):
        degrees = int((dms[0][0] / dms[0][1] +
                       dms[1][0] / dms[1][1] / 60.0 +
                       dms[2][0] / dms[2][1] / 3600.0) * 1e7)
        if hemisphere in ("S", "W"):
            degrees = -degrees
        return degrees



import io
import os.path
import setuptools
from setuptools.command.test import test as TestCommand

setup_ver = {}
with open('apicrud/_version.py') as file:
    exec(file.read(), setup_ver)

__long_desc__ = io.open(os.path.join(os.path.dirname(
    os.path.abspath(__file__)), 'README.md'), encoding='utf-8').read()


class PyTest(TestCommand):
    def finalize_options(self):
        TestCommand.finalize_options(self)
        self.test_args = [
            '--junitxml', 'tests/test-result.xml',
            '--ignore', 'media',
            '--cov-report', 'term-missing',
            '--cov-report', 'html',
            '--cov-report', 'xml']
        self.test_suite = True

    def run_tests(self):
        import pytest
        pytest.main(self.test_args)


setuptools.setup(
    version=setup_ver.get('__version__'),
    name='apicrud',
    description='Flask REST framework for role-based access',
    long_description=__long_desc__,
    long_description_content_type='text/markdown',
    keywords='api flask rbac rest',
    author='Rich Braun',
    author_email='docker@instantlinux.net',
    url='https://github.com/instantlinux/apicrud',
    license='Apache License Version 2.0',
    scripts=[],
    package_data=dict(apicrud=['*.yaml', 'i18n/*/LC_MESSAGES/*.mo',
                               'templates/*.j2']),
    packages=('apicrud', 'apicrud.media', 'apicrud.messaging'),
    include_package_data=True,
    install_requires=[
        'alembic>=1.4.3',
        'babel>=2.8.0',
        'connexion>=2.7.0',
        'connexion[swagger-ui]',
        'cryptography>=2.1.4',
        'Flask>=1.1.2',
        'Flask-Babel>=2.0.0',
        'Flask-Cors>=3.0.8',
        'geocoder>=1.38.1',
        'itsdangerous>=1.1.0',
        'PyJWT>=1.7.1',
        'passlib>=1.7.4',
        'pycryptodomex>=3.9.8',
        'PyMySQL>=0.10.1',
        'pytz>=2020.5',
        'redis>=3.5.3',
        'requests>=2.23.0',
        'SQLAlchemy>=1.3.22',
        'SQLAlchemy-Utils>=0.36.8',
        'swagger-ui-bundle>=0.0.8',
        'urllib3>=1.25.9'],
    python_requires='>=3.6',
    test_suite='tests',
    cmdclass={'test': PyTest},
    classifiers=[
        'Development Status :: 3 - Alpha',
        'License :: OSI Approved :: Apache Software License',
        'Topic :: Software Development :: Libraries :: Application Frameworks',
        'Intended Audience :: Developers',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8'
    ]
)



"""session_auth

created 26-mar-2020 by richb@instantlinux.net
"""
from datetime import datetime, timedelta
from flask import g, request
from flask_babel import _
import jwt
import logging
from passlib.hash import sha256_crypt
from redis.exceptions import ConnectionError
from sqlalchemy import or_
from sqlalchemy.exc import OperationalError
from sqlalchemy.orm.exc import NoResultFound
from sqlalchemy_utils.types.encrypted.padding import InvalidPaddingError
import string
import time

from .access import AccessControl
from .const import Constants
from .messaging.confirmation import Confirmation
from .service_config import ServiceConfig
from .service_registry import ServiceRegistry
from .utils import gen_id, utcnow

APIKEYS = {}


class SessionAuth(object):
    """Session Authorization

    Functions for login, password and role authorization

    Args:
      func_send(function): name of function for sending message
    """

    def __init__(self, func_send=None):
        config = self.config = ServiceConfig().config
        self.models = ServiceConfig().models
        self.jwt_secret = config.JWT_SECRET
        self.login_attempts_max = config.LOGIN_ATTEMPTS_MAX
        self.login_lockout_interval = config.LOGIN_LOCKOUT_INTERVAL
        self.login_session_limit = config.LOGIN_SESSION_LIMIT
        self.login_admin_limit = config.LOGIN_ADMIN_LIMIT
        self.func_send = func_send

    def account_login(self, username, password, roles_from=None):
        """ Log in with username or email

        Args:
          username (str): account name or email
          password (str): credential
          identity_from (obj):  model from which to find contact info
          roles_from (obj): model for which to look up authorizations

        Returns:
          dict:
            Fields include jwt_token (contains uid / account ID),
            ID of entry in settings database, and a sub-dictionary
            with mapping of endpoints registered to microservices
        """

        try:
            if '@' in username:
                account = g.db.query(self.models.Account).join(
                    self.models.Person).filter(
                        self.models.Person.identity == username,
                        self.models.Account.status == 'active').one()
                username = account.name
            else:
                account = g.db.query(self.models.Account).filter_by(
                    name=username, status='active').one()
        except InvalidPaddingError:
            logging.error('action=login message="invalid DB_AES_SECRET"')
            return dict(message=_(u'DB operational error')), 503
        except NoResultFound:
            return dict(username=username, message='not valid'), 403
        except OperationalError as ex:
            logging.error('action=login message=%s' % str(ex))
            return dict(message=_(u'DB operational error, try again')), 403
        if (account.invalid_attempts >= self.login_attempts_max and
            account.last_invalid_attempt + timedelta(
                seconds=self.login_lockout_interval) > datetime.utcnow()):
            time.sleep(5)
            return dict(username=username, message=_(u'locked out')), 403
        if account.password == '':
            logging.error("username=%s, message='no password'" % username)
            return dict(username=username, message=_(u'no password')), 403
        elif not sha256_crypt.verify(password, account.password):
            account.invalid_attempts += 1
            account.last_invalid_attempt = datetime.utcnow()
            logging.info(
                'action=login username=%s credential=invalid attempt=%d' %
                (username, account.invalid_attempts))
            g.db.commit()
            return dict(username=username, message=_(u'not valid')), 403
        logging.info('action=login username=%s account_id=%s' %
                     (username, account.id))
        account.last_login = datetime.utcnow()
        account.invalid_attempts = 0
        g.db.commit()
        # connexion doesn't support cookie auth. probably just as well,
        #  this forced me to implement a better JWT design with nonce token
        # https://github.com/zalando/connexion/issues/791
        #  TODO come up with a standard approach, this is still janky
        #  and does not do signature validation (let alone PKI)

        # TODO research concerns raised in:
        # https://paragonie.com/blog/2017/03/jwt-json-web-tokens-is-bad-
        #  standard-that-everyone-should-avoid
        if (account.password_must_change):
            roles = ['person', 'pwchange']
        # TODO parameterize model name contact
        elif (g.db.query(self.models.Contact).filter(
                self.models.Contact.info == account.owner.identity,
                self.models.Contact.type == 'email'
                ).one().status == 'active'):
            roles = ['admin', 'user'] if account.is_admin else ['user']
            if roles_from:
                roles += self.get_roles(account.uid, roles_from)
        else:
            roles = []
        duration = (self.login_admin_limit if account.is_admin
                    else self.login_session_limit)
        ses = g.session.create(account.uid, roles, acc=account.id,
                               identity=account.owner.identity, ttl=duration)
        if ses:
            retval = dict(
                jwt_token=jwt.encode(
                    ses, self.jwt_secret, algorithm='HS256').decode('utf-8'),
                resources=ServiceRegistry().find()['url_map'],
                settings_id=account.settings_id)
            if hasattr(account.settings, 'default_storage_id'):
                retval['storage_id'] = account.settings.default_storage_id
            return retval, 201
        else:
            return dict(message=_(u'DB operational error')), 500

    def api_access(self, apikey, roles_from=None):
        """ Access using API key

        Args:
          apikey (str): the API key
          roles_from (obj): model for which to look up authorizations

        Returns:
          dict: uid, scopes (None if not authorized)
        """
        global APIKEYS
        acc = AccessControl()
        key_id, secret = apikey.split('.')
        if key_id not in APIKEYS or utcnow() > APIKEYS[key_id]['expires']:
            # Note - local caching in APIKEYS global var reduces database
            # queries on Accounts table
            uid, scopes = acc.apikey_verify(key_id, secret)
            if not uid:
                return None
            try:
                account = g.db.query(self.models.Account).filter_by(
                    uid=uid, status='active').one()
            except NoResultFound:
                logging.info(dict(action='api_key', uid=uid,
                                  message='account not active'))
                return None
            except Exception as ex:
                logging.error(dict(action='api_key', message=str(ex)))
                return None
            APIKEYS[key_id] = dict(
                account_id=account.id,
                expires=utcnow() + timedelta(seconds=self.config.REDIS_TTL),
                identity=account.owner.identity,
                roles=['admin', 'user'] if account.is_admin else ['user'],
                scopes=[item.name for item in scopes], uid=uid)
            if roles_from:
                APIKEYS[key_id]['roles'] += self.get_roles(uid, roles_from)
        item = APIKEYS[key_id]
        uid = item['uid']
        logging.debug(dict(action='api_key', uid=uid, scopes=item['scopes']))
        duration = (self.login_admin_limit if 'admin' in item['roles']
                    else self.login_session_limit)
        token = acc.apikey_hash(secret)[:8]
        ses = None
        try:
            # Look for existing session in redis cache
            ses = g.session.get(
                uid, token, arg='auth', key_id=key_id).split(':')
        except AttributeError:
            ses = g.session.create(uid, item['roles'], acc=item['account_id'],
                                   identity=item['identity'],
                                   ttl=duration, key_id=key_id, nonce=token)
        if ses:
            return dict(uid=uid, scopes=item['scopes'])

    def register(self, identity, username, name, template='confirm_new'):
        """Register a new account: create related records in database
        and send confirmation token to new user

        TODO caller still has to invoke account-create function to
        generate record in accounts table

        Args:
          identity (str): account's primary identity, usually an email
          username (str): account's username
          name (str): name
          template (str): template for message (confirming new user)
        Returns:
          tuple: the Confirmation.request dict and http response
        """
        if not username or not identity or not name:
            return dict(message=_(u'all fields required')), 400
        identity = identity.lower()
        logmsg = dict(action='register', identity=identity,
                      username=username)
        try:
            g.db.query(self.models.Account).filter_by(name=username).one()
            msg = _(u'that username is already registered')
            logging.warning(dict(message=msg, **logmsg))
            return dict(message=msg), 405
        except NoResultFound:
            pass
        uid = None
        try:
            existing = g.db.query(self.models.Person).filter_by(
                identity=identity).one()
            uid = existing.id
            g.db.query(self.models.Account).filter_by(uid=uid).one()
            msg = _(u'that email is already registered, use forgot-password')
            logging.warning(dict(message=msg, **logmsg))
            return dict(message=msg), 405
        except NoResultFound:
            pass
        if uid:
            try:
                cid = g.db.query(self.models.Contact).filter_by(
                    info=identity, type='email').one().id
            except Exception as ex:
                msg = 'registration trouble, error=%s' % str(ex)
                logging.error(dict(message=msg, **logmsg))
                return dict(message=msg), 405
        else:
            person = self.models.Person(
                id=gen_id(prefix='u-'), name=name, identity=identity,
                status='active')
            uid = person.id
            cid = gen_id()
            g.db.add(person)
            g.db.add(self.models.Contact(id=cid, uid=uid, type='email',
                                         info=identity))
            g.db.commit()
            logging.info(dict(message=_(u'person added'), uid=uid, **logmsg))
        # If browser language does not match global default language, add
        # a profile setting
        lang = request.accept_languages.best_match(self.config.LANGUAGES)
        try:
            default_lang = g.db.query(self.models.Settings).filter_by(
                name='global').one().lang
        except Exception:
            default_lang = 'en'
        if lang and lang != default_lang:
            g.db.add(self.models.Profile(id=gen_id(), uid=uid, item='lang',
                                         value=lang, status='active'))
            g.db.commit()
        return Confirmation().request(cid, template=template,
                                      func_send=self.func_send)

    def change_password(self, uid, new_password, reset_token,
                        old_password=None, verify_password=None):
        """Update a user's password, applying complexity rules; must
        specify either the old password or a reset token

        Args:
          uid (str): User ID
          new_password (str): the new passphrase
          reset_token (str): a token retrieved from Confirmation.request
          old_password (str): the old passphrase

        Returns:
          tuple: dict with account_id/uid/username, http response
        """

        if not new_password or new_password != verify_password:
            return dict(message=_(u'passwords do not match')), 405
        try:
            account = g.db.query(self.models.Account).filter_by(
                uid=uid, status='active').one()
        except NoResultFound:
            return dict(uid=uid, message=_(u'account not found')), 404
        logmsg = dict(action='change_password', resource='account',
                      username=account.name, uid=uid)

        if self._password_weak(new_password):
            return dict(message=_(u'rejected weak password')), 405
        if old_password:
            if not sha256_crypt.verify(old_password, account.password):
                msg = 'invalid credential'
                logging.warning(dict(message=msg, **logmsg))
                return dict(username=account.name, message=msg), 403
        else:
            retval = Confirmation().confirm(reset_token)
            if retval[1] != 200:
                msg = _(u'invalid token')
                logging.warning(dict(message=msg, **logmsg))
                return dict(username=account.name, message=msg), 405
        account.password = sha256_crypt.hash(new_password)
        account.password_must_change = False
        g.db.commit()
        logging.info(dict(message=_(u'changed'), **logmsg))
        return dict(id=account.id, uid=uid, username=account.name), 200

    def forgot_password(self, identity, username, template='password_reset'):
        """Trigger Confirmation.request; specify either the username
        or email address

        Args:
          identity (str): account's primary identity, usually an email
          username (str): account's username
          template (str): template for message (confirming new user)
        Returns:
          tuple: the Confirmation.request dict and http response
        """

        logmsg = dict(action='forgot_password', identity=identity,
                      username=username)
        try:
            account = g.db.query(self.models.Account).join(
                self.models.Person).filter(
                    self.models.Account.is_admin == 0,
                    self.models.Account.status == 'active',
                or_(
                    self.models.Account.name == username,
                    self.models.Person.identity == identity)).one()
            id = g.db.query(self.models.Contact).filter_by(
                type='email', info=account.owner.identity).one().id
        except NoResultFound:
            logging.warning(dict(message='not found', **logmsg))
            if identity and '@' in identity:
                # TODO send reset_invalid message instead of rejection
                return dict(message=_(u'username or email not found')), 404
            return dict(message=_(u'username or email not found')), 404
        logging.info(logmsg)
        return Confirmation().request(
            id, template=template, func_send=self.func_send)

    def get_roles(self, uid, member_model, resource=None, id=None):
        """Get roles that match uid / id for a resource
        Each is in the form <resource>-<id>-<privacy level>

        Args:
          uid (str): User ID
          member_model (obj): the DB model that defines membership in resource
          resource (str): the resource that defines privacy (e.g. list)
          id (str): ID of the resource (omit if all are desired)
        Returns:
          list of str: authorized roles
        """
        acc = AccessControl()
        if not resource:
            resource = acc.primary_res
        column = '%s_id' % resource
        try:
            query = g.db.query(member_model).filter_by(uid=uid,
                                                       status='active')
            if id:
                query = query.filter_by(**{column: id})
        except Exception as ex:
            logging.warning('action=get_roles message="%s"' % str(ex))
        roles = []
        for record in query.all():
            if not hasattr(record, resource):
                continue
            # Skip first defined level (always 'public')
            for level in acc.privacy_levels[1:]:
                roles.append('%s-%s-%s' % (resource, getattr(record, column),
                                           level))
                if level == record.authorization:
                    break
        return roles

    def update_auth(self, member_model, id, resource=None, force=False):
        """Check current access, update if recently changed

        Args:
          member_model (obj): model (e.g. Guest) which defines membership in
            resource
          id (str): resource id of parent resource
          resource (str):
            parent resource for which membership should be checked
          force (bool): perform update regardless of logged-in permissions
        """
        acc = AccessControl()
        if not resource:
            resource = acc.primary_res
        current = set(acc.auth)
        if force or ('%s-%s-%s' % (resource, id, Constants.AUTH_INVITEE)
                     not in current):
            # TODO handle privilege downgrade member/host->invitee
            current |= set(self.get_roles(acc.uid, member_model,
                                          resource=resource, id=id))
            creds = request.authorization
            g.session.update(creds.username, creds.password, 'auth',
                             ':'.join(current))

    @staticmethod
    def _password_weak(password):
        """Validate password complexity

        Args:
          password (str): Password
        Returns:
          bool: True if at least 3 different types of characters
        """
        return (not set(password).intersection(string.ascii_lowercase),
                not set(password).intersection(string.ascii_uppercase),
                not set(password).intersection(string.digits),
                not set(password).intersection(string.punctuation)
                ).count(True) > 1


# Support openapi securitySchemes: basic_auth and api_key endpoints

def basic(username, password, required_scopes=None):
    """This is a modified basic-auth validation function. The auth login
    controller method generates a random 8-byte token, stores it in
    the session_manager object as token_auth, and sends it to javascript
    authProvider. The dataProvider must send it back to us as
    basic-auth (base64-encoded).

    Vulnerable to session-hijacking if auth packets aren't encrypted
    end to end, but "good enough" until OAuth2 effort is completed.

    Implemented because of https://github.com/zalando/connexion/issues/791

    Args:
      username (str): Session UID
      password (str): Session token
      required_scopes (list): not used
    Returns:
      dict: uid with the username passed in
    """

    session = g.session.get(username, password)
    token_auth = 'ok'
    if session is None or 'jti' not in session:
        token_auth = 'missing'
    elif session['jti'] != password or session['sub'] != username:
        token_auth = 'invalid'
    elif 'exp' in session and datetime.utcnow().strftime(
            '%s') > session['exp']:
        token_auth = 'expired'

    if token_auth != 'ok':
        logging.info('action=logout username=%s token_auth=%s' % (
            username, token_auth))
        if session:
            try:
                g.session.delete(username, password)
            except ConnectionError as ex:
                logging.error(dict(action='logout', message=str(ex)))
        return None
    return dict(uid=username)


def api_key(apikey, required_scopes=None):
    """ API key authentication

    Args:
      apikey (str): the key
      required_scopes (list): permissions requested

    Returns:
      dict: uid if successful (None otherwise)
    """
    models = ServiceConfig().models
    if len(apikey) > 96 or '.' not in apikey:
        return None
    retval = SessionAuth().api_access(apikey, roles_from=models.List)
    if not retval:
        return retval
    if required_scopes and (set(required_scopes) & set(retval['scopes']) !=
                            set(required_scopes)):
        logging.info(dict(action='api_key', prefix=apikey.split('.')[0],
                          scopes=retval['scopes'], message='denied'))
        return None
    return retval
